/**
 *
 * @authors yutent (yutent@doui.cc)
 * @date    2017-04-14 21:04:50
 *
 */
import { bind, fire } from 'wkit'
import { hideProp, targetIsThisWindow } from '../utils.js'

//hash前缀正则
const PREFIX_REGEXP = /^(#!|#)[\/]?/
const TRIM_REGEXP = /(^[/]+)|([/]+$)/g
const DEFAULT_OPTIONS = {
  prefix: '!/',
  allowReload: true //连续点击同一个链接是否重新加载
}

const RULE_REGEXP =
  /(:id)|(\{id\})|(\{id:([A-z\d\,\[\]\{\}\-\+\*\?\!:\^\$]*)\})/g

class Router {
  #tables = new Map()
  #views = new Set()

  #targets = new Set()

  #ready = false
  #options = Object.create(null)
  #route = Object.create(null)

  constructor(options = {}) {
    Object.assign(this.#options, DEFAULT_OPTIONS, options)

    bind(window, 'load, popstate', this.#hashchange.bind(this))
  }

  get route() {
    return this.#route
  }

  #hashchange(ev) {
    if (ev?.type === 'load') {
      if (this.#ready) {
        return
      }
      this.#ready = true
    }

    let path = location.hash

    path = path.replace(PREFIX_REGEXP, '').trim()
    path = path.replace(TRIM_REGEXP, '')

    if (ev?.type === 'load') {
      this.go(path)
      // hash模式要手动触发一下路由检测
      this.#check(path)
    } else {
      this.#check(path)
    }
  }

  #parseRule(route) {
    if (route.path === '!') {
      route.regexp = null
    } else {
      let re = route.path.replace(RULE_REGEXP, function (m, p1, p2, p3, p4) {
        let w = '([\\w.-]'
        if (p1 || p2) {
          return w + '+)'
        } else {
          if (!/^\{[\d\,]+\}$/.test(p4)) {
            w = '('
          }
          return w + p4 + ')'
        }
      })
      re = re
        .replace(/(([^\\])([\/]+))/g, '$2\\/')
        .replace(/(([^\\])([\.]+))/g, '$2\\.')
        .replace(/(([^\\])([\-]+))/g, '$2\\-')
        .replace(/(\(.*)(\\[\-]+)(.*\))/g, '$1-$3')
      re = '^' + re + '$'
      route.regexp = new RegExp(re)
    }
    return route
  }

  #add(route) {
    if (route.path !== '!' && route.path[0] !== '/') {
      console.error('路由规则必须以"/"开头')
      return
    }
    route.path = route.path.replace(/^[\/]+|[\/]+$|\s+/g, '')

    this.#tables.set(route.path, this.#parseRule(route))
    this.#views.add(route.name)
  }

  // 路由检测
  #check(path) {
    let { allowReload } = this.#options
    let $view = window.__wkitd__.get('ROUTER_VIEW')

    if (!$view || (!allowReload && path === this.#route.path)) {
      return
    }

    for (let [k, route] of this.#tables) {
      let args = path.match(route.regexp)
      if (args) {
        args.shift()
        $view.current = route.name
        this.#route.path = path
        this.#route.name = route.name
        this.#route.params = args
        this.#route.query = {}
        this.#rsync()
        return
      }
    }
    if (this.#tables.get('!')) {
      route = this.#tables.get('!')
      $view.current = route.name
      this.#route = { path, name: route.name, params: {}, query: {} }
    }
  }

  #rsync() {
    for (let callback of this.#targets) {
      callback(this.route)
    }
  }

  init() {
    this.#hashchange()
  }

  rsync(callback) {
    this.#targets.add(callback)
  }

  getViews() {
    return Array.from(this.#views)
  }

  beforeEach(callback) {
    callback(prev, next, function () {
      //
    })
  }

  // 跳转到路由
  go(path) {
    path = path.trim().replace(TRIM_REGEXP, '')

    // 页面刷新时, 不主动添加空hash, 避免执行2次noMatch回调
    if (!path && path === location.hash) {
      return
    }
    location.hash = '!/' + path
  }

  // 绑定路由事件
  addRoute(routes) {
    if (Array.isArray(routes)) {
      routes.forEach(it => {
        this.#add(it)
      })
    } else {
      this.#add(routes)
    }
    // 因为先初始化,才开始监听路由规则
    // 所以会导致wondow load的时候, 规则还没生效, 而生效之后,load已经结束
    // 所以这里需要手动再触发一次load
    if (this.#ready) {
      this.#hashchange()
    } else {
      fire(window, 'load')
    }
  }
}

export default function () {
  return new Router()
}
